/*
Copyright 2011 Daniel Morton. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY Daniel Morton ''AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those of the
authors and should not be interpreted as representing official policies, either expressed
or implied, of Daniel Morton.
*/

package ca.quadrilateral.wallgenerator;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.util.HashMap;
import java.util.Map;

import javax.media.jai.PerspectiveTransform;

public class Generator {
	private static final float NO_DARKENING_THRESHOLD = 0.009f;
	
    private final Config config;
    private final double nearDarkening;
    private final double farDarkening;

    public Generator(Config config, double nearDarkening, double farDarkening) {
        this.config = config;
        this.nearDarkening = nearDarkening;
        this.farDarkening = farDarkening;
    }

    public Map<String, BufferedImage> generate(BufferedImage sourceImage) {
        final Map<String, BufferedImage> resultMap = new HashMap<String, BufferedImage>();

        final BufferedImage baseImage = scaleImage(sourceImage, "CloseFront");

        resultMap.put("CloseFront", baseImage);

        final Dimension closeFrontSideSize = config.getWallSize("CloseFrontSide");
        final Dimension nearFrontSideSize = config.getWallSize("NearFrontSide");

        final BufferedImage baseImageRightHalf = baseImage.getSubimage(baseImage.getWidth() / 2, 0, baseImage.getWidth() / 2, baseImage.getHeight());
        final BufferedImage closeFrontRight = baseImage.getSubimage(baseImage.getWidth() - closeFrontSideSize.width, 0, closeFrontSideSize.width, baseImage.getHeight());

        resultMap.put("CloseFrontRight", closeFrontRight);

        final BufferedImage baseImageLeftHalf = baseImage.getSubimage(0, 0, baseImage.getWidth() / 2, baseImage.getHeight());
        final BufferedImage closeFrontLeft = baseImage.getSubimage(0, 0, closeFrontSideSize.width, baseImage.getHeight());

        resultMap.put("CloseFrontLeft", closeFrontLeft);

        BufferedImage nearByFront = scaleImage(baseImage, "NearFront");

        final BufferedImage nearByFrontOriginal = new BufferedImage(nearByFront.getWidth(), nearByFront.getHeight(), BufferedImage.TYPE_INT_ARGB);

        final Graphics2D nearByFrontOriginalGraphics = nearByFrontOriginal.createGraphics();
        nearByFrontOriginalGraphics.drawImage(nearByFront, 0, 0, null);
        nearByFrontOriginalGraphics.dispose();
        
        nearByFront = getDarkenedSourceImage(nearByFront, (float)nearDarkening, (float)nearDarkening, true);
        
        final BufferedImage nearByFrontRight = nearByFront.getSubimage(nearByFront.getWidth() - nearFrontSideSize.width, 0, nearFrontSideSize.width, nearByFront.getHeight());
        final BufferedImage nearByFrontLeft = nearByFront.getSubimage(0, 0, nearFrontSideSize.width, nearByFront.getHeight());

        resultMap.put("NearFront", nearByFront);
        resultMap.put("NearFrontRight", nearByFrontRight);
        resultMap.put("NearFrontLeft", nearByFrontLeft);

        BufferedImage farFront = scaleImage(nearByFront, "FarFront");
        
        final BufferedImage farFrontOriginal = new BufferedImage(farFront.getWidth(), farFront.getHeight(), BufferedImage.TYPE_INT_ARGB);
        
        final Graphics2D farFrontOriginalGraphics = farFrontOriginal.createGraphics();
        farFrontOriginalGraphics.drawImage(farFront, 0, 0, null);
        farFrontOriginalGraphics.dispose();
                
        farFront = getDarkenedSourceImage(farFront, (float)farDarkening, (float)farDarkening, true);
        
        
        
        resultMap.put("FarFront", farFront);

        resultMap.put("FarFrontSecondary", farFront);

        final BufferedImage closeLeft = generateTransformedImage(baseImageRightHalf, "CloseSide", "CloseLeft");
        resultMap.put("CloseLeft", closeLeft);

        final BufferedImage closeRight = generateTransformedImage(baseImageLeftHalf, "CloseSide", "CloseRight");
        resultMap.put("CloseRight", closeRight);

        final BufferedImage nearLeft = generateTransformedImage(getDarkenedSourceImage(baseImage, 0f, (float)nearDarkening, true), "NearSide", "NearLeft");
        resultMap.put("NearLeft", nearLeft);

        final BufferedImage nearRight = generateTransformedImage(getDarkenedSourceImage(baseImage, 0f, (float)nearDarkening, false), "NearSide", "NearRight");
        resultMap.put("NearRight", nearRight);

        final BufferedImage farLeft = generateTransformedImage(getDarkenedSourceImage(nearByFrontOriginal, (float)nearDarkening, (float)farDarkening, true), "FarSide", "FarLeft");
        resultMap.put("FarLeft", farLeft);

        final BufferedImage farRight = generateTransformedImage(getDarkenedSourceImage(nearByFrontOriginal, (float)nearDarkening, (float)farDarkening, false), "FarSide", "FarRight");
        resultMap.put("FarRight", farRight);

        return resultMap;
    }
    
    
    
    private BufferedImage getDarkenedSourceImage(BufferedImage sourceImage, float startAlpha, float endAlpha, boolean leftToRight) {
    	if (startAlpha > NO_DARKENING_THRESHOLD || endAlpha > NO_DARKENING_THRESHOLD) {
    		final BufferedImage blackSource = new BufferedImage(1, sourceImage.getHeight(), BufferedImage.TYPE_INT_ARGB);
    		final BufferedImage modifiedSourceImage = new BufferedImage(sourceImage.getWidth(), sourceImage.getHeight(), BufferedImage.TYPE_INT_ARGB);
    		
    		final Graphics2D blackSourceGraphics = blackSource.createGraphics();
    		blackSourceGraphics.setColor(Color.BLACK);
    		blackSourceGraphics.fillRect(0, 0, 1, sourceImage.getHeight());
    		blackSourceGraphics.dispose();
    
    		final Graphics2D graphics = modifiedSourceImage.createGraphics();
    		graphics.drawImage(sourceImage, 0, 0, null);
    		
    		final float delta = (endAlpha - startAlpha) / sourceImage.getWidth();
    		
    		if (leftToRight) {
    			for(int i = 0; i < sourceImage.getWidth(); i++) {
    				final float alpha = startAlpha + (delta * i);
    				graphics.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, alpha));
    				graphics.drawImage(blackSource, i, 0, null);
    			}
    		} else {
    			for(int i = sourceImage.getWidth() - 1; i >= 0; i--) {
    				final float alpha = startAlpha + (delta * ((sourceImage.getWidth() - 1) - i));
    				graphics.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, alpha));
    				graphics.drawImage(blackSource, i, 0, null);    				
    			}
    		}    		    		    		
    		graphics.dispose();
    		
    		for(int x = 0; x < sourceImage.getWidth(); x++) {
    			for(int y = 0; y < sourceImage.getHeight(); y++) {
    				int sourceAlphaValue = (sourceImage.getRGB(x, y) & 0xFF000000);
    				int targetPixelColor = (modifiedSourceImage.getRGB(x, y) & 0x00FFFFFF);
    				modifiedSourceImage.setRGB(x, y, (targetPixelColor | sourceAlphaValue));
    			}
    		}
    		
    		return modifiedSourceImage;
    	} else {
    		return sourceImage;
    	}
    }

    private BufferedImage generateTransformedImage(BufferedImage sourceImage, String wallSizeKey, String transformKey) {
        final PerspectiveTransform transform = config.getTransform(transformKey);

        final Dimension wallSize = config.getWallSize(wallSizeKey);

        final BufferedImage resultImage = new BufferedImage(wallSize.width, wallSize.height, BufferedImage.TYPE_INT_ARGB);

        applyTransform(sourceImage, resultImage, transform);
        return resultImage;
    }

    private BufferedImage scaleImage(BufferedImage sourceImage, String configKey) {
        final Dimension wallSize = config.getWallSize(configKey);
        final BufferedImage image = new BufferedImage(wallSize.width, wallSize.height, BufferedImage.TYPE_INT_ARGB);

        final Graphics2D graphics = image.createGraphics();

        graphics.drawImage(sourceImage, 0, 0, wallSize.width, wallSize.height, null);

        graphics.dispose();

        return image;
    }

    private void applyTransform(BufferedImage sourceImage, BufferedImage destinationImage, PerspectiveTransform transform) {
        for(int x = 0; x < destinationImage.getWidth(); x++) {
            for(int y = 0; y < destinationImage.getHeight(); y++) {
                try {
                    final Point2D sourcePoint = new Point2D.Double(x, y);
                    final Point2D destPoint = new Point2D.Double();
                    transform.inverseTransform(sourcePoint, destPoint);

                    int destPointX = (int)destPoint.getX();
                    int destPointY = (int)destPoint.getY();

                    if (destPointX < 0 ||
                            destPointY < 0 ||
                            destPointX >= sourceImage.getWidth() ||
                            destPointY >= sourceImage.getHeight()) {
                        continue;
                    }

                    final int color = sourceImage.getRGB(destPointX, destPointY);
                    destinationImage.setRGB((int)sourcePoint.getX(), (int)sourcePoint.getY(), color);
                } catch (Exception e) {
                    throw new RuntimeException("Error transforming image!", e);
                }
            }
        }
    }
}
